import Flutter
import UIKit
import CoreLocation

public class SwiftAjlocationPlugin: NSObject, FlutterPlugin {
    
    private let locationManager = CLLocationManager()
    
    private var eventChannelLocationUpdateSink: FlutterEventSink?
    
    private var methodChannelResult: FlutterResult?
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "ajlocation", binaryMessenger: registrar.messenger())
        let eventChannel = FlutterEventChannel(name: "ajlocation.locationChange", binaryMessenger: registrar.messenger())
        let instance = SwiftAjlocationPlugin()
        eventChannel.setStreamHandler(instance)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        methodChannelResult = result
        if call.method == "requestAuthorization" {
            guard let type = call.arguments as? Int else {
                return
            }
            NSLog("\n \(type)")
            requestAuthorization(type: type)
        } else if call.method == "authorizationStatus" {
            currentAuthorizationStatus()
        } else if call.method == "locationMangaerConfiguration" {
            guard let config = call.arguments as? [String: Any] else {
                return
            }
            setLocationManagerConfiguration(configuration: config)
        } else if call.method == "currentLocation" {
            getCurrentLocation()
        } else if call.method == "startUpdatingLocation" {
            startUpdatingLocation()
        } else if call.method == "stopUpdatingLocation" {
            stopUpdatingLocation()
        } else if call.method == "setConfiguration" {
            guard let configuration = call.arguments as? [String: Any] else {
                return
            }
            setLocationManagerConfiguration(configuration: configuration)
        }
    }
}

extension SwiftAjlocationPlugin: CLLocationManagerDelegate, FlutterStreamHandler {
    
    //MARK: - FlutterStreamHandler
    
    public func onListen(withArguments arguments: Any?, eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        self.eventChannelLocationUpdateSink = events
        return nil
    }
    
    public func onCancel(withArguments arguments: Any?) -> FlutterError? {
        self.eventChannelLocationUpdateSink = nil
        return nil
    }
    
    
    //MARK: - CLLocationManager
    
    private func requestAuthorization(type: Int) {
        locationManager.delegate = self
        if type == 0 {
            locationManager.requestAlwaysAuthorization()
            locationManager.allowsBackgroundLocationUpdates = true
        } else if type == 1 {
            locationManager.requestWhenInUseAuthorization()
        } else {
            return
        }
    }
    
    private func currentAuthorizationStatus() {
        switch CLLocationManager.authorizationStatus() {
        case .authorizedAlways:
            methodChannelResult?(0)
            break
        case .authorizedWhenInUse:
            methodChannelResult?(1)
            break
        case .denied:
            methodChannelResult?(2)
            break
        case .notDetermined:
            methodChannelResult?(3)
            break
        case .restricted:
            methodChannelResult?(4)
            break
        }
    }
    
    private func setLocationManagerConfiguration(configuration: [String: Any]) {
        guard let distance = configuration["distanceFilter"] as? Double,
            let accuracy = configuration["accuracy"] as? Int else {
                return
        }
        
        switch accuracy {
        case 0:
            locationManager.desiredAccuracy = kCLLocationAccuracyBestForNavigation
            break
        case 1:
            locationManager.desiredAccuracy = kCLLocationAccuracyBest
            break
        case 2:
            locationManager.desiredAccuracy = kCLLocationAccuracyKilometer
            break
        case 3:
            locationManager.desiredAccuracy = kCLLocationAccuracyHundredMeters
            break
        case 4:
            locationManager.desiredAccuracy = kCLLocationAccuracyThreeKilometers
            break
        case 5:
            locationManager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters
            break
        default:
            locationManager.desiredAccuracy = kCLLocationAccuracyBest
            break
        }
        locationManager.distanceFilter = distance
    }
    
    private func startUpdatingLocation() {
        locationManager.startUpdatingLocation()
    }
    
    private func stopUpdatingLocation() {
        locationManager.stopUpdatingLocation()
    }
    
    private func getCurrentLocation() {
        guard let location = locationManager.location else {
            return
        }
        
        let dict = createLocationDictionary(location)
        methodChannelResult?(dict)
    }
    
    public func locationManager(_ manager: CLLocationManager, didUpdateLocations locations: [CLLocation]) {
        guard let location = manager.location else {
            return
        }
        
        let dict = createLocationDictionary(location)
        eventChannelLocationUpdateSink?(dict)
    }
    
    public func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {
        currentAuthorizationStatus()
    }
}

extension SwiftAjlocationPlugin {
    
    private func createLocationDictionary(_ location: CLLocation) -> [String: Any] {
        return [
            "altitude": location.altitude,
            "longitude": location.coordinate.longitude,
            "latitude": location.coordinate.latitude,
            "accuracy": location.horizontalAccuracy,
            "timestamp": location.timestamp.timeIntervalSince1970,
            "speed": location.speed
        ]
    }
}
